//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtmHandler.c *#*#*#*#*#*#*#*# (C) 2012-2019 DekTec
//
// DtmHandler - Protocol handler for DTM-32xx devices - Implementation
//
// The open source "DTM handler" is a set of definitions and functions that can help you 
// build a protocol handler for DekTec's DTM-32xx devices. This C source code can be 
// compiled and linked into a C or C++ application. Please refer to DtmCmd for an open
// source example of a DTM-32xx command-line control program that uses the DTM handler.
//
// This file contains the implementation of the DTM handler functions. Definitions and
// declarations that are used internally in the DTM handler are defined in this file,
// while public definitions required by applications that use the DTM handler are
// contained in DtmHandler.h
//
// Please refer to DtmHandler.h for further usage information of the DTM handler.

//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ LICENSE +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Copyright (C) 2012-2019 DekTec Digital Video B.V.
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this list
//     of conditions and the following disclaimer.
//  2. Redistributions in binary format must reproduce the above copyright notice, this
//     list of conditions and the following disclaimer in the documentation.
//  3. The source code may not be modified for the express purpose of enabling hardware
//     features for which no genuine license has been obtained.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DEKTEC DIGITAL VIDEO BV, ITS AGENTS OR ITS EMPLOYEES BE LIABLE FOR
// ANY DIRECT, INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR OTHER DAMAGES (INCLUDING DAMAGES
// FOR THE LOSS OF USE, INFORMATION, GOODWILL, PROFIT, WORK STOPPAGE, DATA, BUSINESS OR
// REVENUE) UNDER ANY CIRCUMSTANCES, OR UNDER ANY LEGAL THEORY, WHETHER IN CONTRACT, IN
// TORT, IN NEGLIGENCE, OR OTHERWISE, ARISING FROM THE USE OF, OR INABILITY TO USE THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Includes -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#include "DtmHandler.h"             // Public header file for the DTM handler

#ifdef USE_DYN_UPG_BUFFER
#include <stdlib.h>                 // Required for malloc() and free() functions
#endif

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Type Definitions -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

#ifndef FALSE
    #define  FALSE 0
#endif
#ifndef TRUE
    #define  TRUE  1
#endif

#ifndef NULL
    #define  NULL  ((void *)0)
#endif

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmMessage -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Data structure describing a DTM protocol message. It is used to store incoming 
// messages in a parsed format, and to specify a message to be transmitted
//
typedef struct _DtmMessage
{
    int  m_Address;                 // Address of the DTM-32xx that is targeted
    int  m_Category;                // DTM protocol category number
    int  m_Setting;                 // DTM protocol setting number
    int  m_ReadWrite;               // DTM protocol read/write bit
    int  m_Index;                   // DTM protocol index number
    int  m_Checksum;                // DTM protocol message checksum
    int  m_DataLength;              // DTM protocol data size in bytes
    int  m_ConvBinToHex;            // 1 = Convert binary message data to ASCII hex
    int  m_ReadBytes;               // 1 = DtmReadBytes is used to read multiple bytes
    int  m_WriteBytes;              // 1 = DtmWriteBytes is used to write multiple bytes
    char*  m_pData;                 // Pointer to the data to be transmitted or received
} DtmMessage;

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Constants -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Protocol definitions
#define DTM_FIELD_RW_READ           1            // Read/write bit for a read command
#define DTM_FIELD_RW_WRITE          0            // Read/write bit for a write command
#define DTM_MAX_DATA_SIZE           8            // Maximum data size for commands that
                                                 // do not use a byte array
#define DTM_MAX_UART_NONDATABYTES   16           // #Protocol bytes for UART
#define DTM_MAX_I2C_NONDATABYTES    7            // #Protocol bytes for I2C

// Protocol-dependent size of file parts when upgrading the DTM-32xx firmware
#define DTM_UPG_UART_PART_SIZE      1024         // Size when upgrading through UART
#define DTM_UPG_I2C_PART_SIZE       200          // Size when upgrading through I2C

// Total size of the upgrade buffer in various modes
#define DTM_UPG_SIZE_ASC            (DTM_MAX_UART_NONDATABYTES                           \
                                                       + (DTM_UPG_UART_PART_SIZE*8/7 + 1))
#define DTM_UPG_SIZE_HEX            (DTM_MAX_UART_NONDATABYTES + DTM_UPG_UART_PART_SIZE*2)
#define DTM_UPG_SIZE_I2C            (DTM_MAX_I2C_NONDATABYTES + DTM_UPG_I2C_PART_SIZE)

// The initialization of a firmware upgrade (sequence of actions before sending the file 
// parts) is performed in a number of steps. This number is defined here to determine the
// "progress increment" used by the callback function to report progress.
#define DTM_NUM_OF_UPG_INIT_STAGES  5
#define DTM_UPG_INIT_PROGRESS_INCR  (100 / DTM_NUM_OF_UPG_INIT_STAGES)

//-.-.-.-.-.-.-.-.-.-.-.-.- DTM-32xx Protocol Description Tables -.-.-.-.-.-.-.-.-.-.-.-.-

// This table defines the settings categories available on DTM-32xx devices and provides
// basic info for each of them
const DtmCategoryInfo  g_DtmCategoryInfo[] =
{
    { 0x01, 0, 300, "Device properties" },
    { 0x02, 0, 300, "Overall configuration" },
    { 0x03, 0, 300, "Network settings" },
    { 0x7F, 0, 5000, "Factory settings" },
    { 0x80, 1, 1100, "Firmware update type 1 (DTM-3200)" },
    { 0x81, 1, 300, "IP receive settings" },
    { 0x82, 1, 1100, "IP transmit settings" },
    { 0x83, 1, 1100, "ASI/parallel input settings" },
    { 0x84, 1, 300, "ASI/parallel output settings" },
    { 0x85, 1, 600, "Firmware upgrade type 2 (DTM-3237)" },
    { 0x86, 1, 600, "Firmware upgrade type 3 (DTM-3224)" },
    { 0x90, 1, 300, "Receiver settings" },
    { 0x91, 1, 300, "Receiver status" },
    { 0x92, 1, 300, "Receiver statistics" },
    { 0x93, 1, 300, "DiSEqC send and receive commands" },
    { 0x94, 1, 300, "Satellite information" }
};
const int  g_NumDtmCategoryInfoItems = sizeof(g_DtmCategoryInfo) 
                                                           / sizeof(g_DtmCategoryInfo[0]);

// This table defines the settings available on DTM-32xx devices and provides basic info
// for each of them
const DtmSettingInfo  g_DtmSettingInfo[] =
{
    { 0x01, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x01, 0x01, 1, D_UINT8, "FPGA version" },
    { 0x01, 0x02, 1, D_UINT8, "Hardware version (DTM-3200)" },
    { 0x01, 0x03, 1, D_UINT32, "Firmware version" },
    { 0x01, 0x04, 1, D_UINT32, "Serial number" },
    { 0x01, 0x05, 1, D_UINT32, "Type" },
    { 0x01, 0x06, 1, D_UINT32, "Hardware version (DTM-3237)" },
    { 0x01, 0x07, 1, D_UINT8, "Bootloader version" },
    { 0x01, 0x08, 1, D_UINT32, "Production date" },
    { 0x01, 0x09, 1, D_UINT8, "I2C address" },
    { 0x01, 0x0A, 1, D_UINT32, "RS-232 baud rate" },
    { 0x01, 0x0B, 1, D_UINT8, "Subtype" },
    { 0x01, 0x0C, 1, D_UINT8, "Busy" },
    { 0x01, 0x0D, 10, D_INT32, "RF level calibration" },
    { 0x02, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x02, 0x01, 1, D_UINT8, "Mode" },
    { 0x02, 0x02, 1, D_UINT8, "Application" },
    { 0x02, 0x03, 1, D_UINT8, "Volatile storage" },
    { 0x02, 0x04, 1, D_UINT8, "Persist all settings" },
    { 0x02, 0x05, 1, D_UINT32, "Number of writes to flash memory" },
    { 0x03, 0x01, 1, D_IP, "IP address" },
    { 0x03, 0x02, 1, D_IP, "Subnet mask" },
    { 0x03, 0x03, 1, D_IP, "Gateway" },
    { 0x03, 0x04, 1, D_UINT8, "DHCP enabled" },
    { 0x03, 0x05, 1, D_UINT8, "Reboot" },
    { 0x03, 0x06, 1, D_MAC, "MAC address" },
    { 0x03, 0x07, 1, D_UINT8, "IP Version" },
    { 0x03, 0x08, 1, D_IPV6, "IPv6 address" },
    { 0x03, 0x09, 1, D_UINT8, "IPv6 prefix" },
    { 0x03, 0x0A, 1, D_IPV6, "IPv6 gateway" },
    { 0x03, 0x0B, 1, D_UINT8, "Port enable" },
    { 0x03, 0x0C, 1, D_UINT8, "Channel settings status" },
    { 0x03, 0x0D, 1, D_UINT8, "Link state" },
    { 0x03, 0x0E, 4, D_UINT32, "Tx count" },
    { 0x03, 0x0F, 4, D_UINT32, "Rx count" },
    { 0x7F, 0x01, 1, D_UINT8, "Factory setting" },
    { 0x7F, 0x02, 1, D_UINT8, "Factory setting" },
    { 0x7F, 0x03, 1, D_UINT8, "Factory setting" },
    { 0x7F, 0x04, 1, D_UINT8, "Factory setting" },
    { 0x7F, 0x05, 1, D_UINT8, "Factory setting" },
    { 0x80, 0x01, 1, D_UINT16, "Part number" },
    { 0x80, 0x02, 1, D_UINT8, "File part" },
    { 0x80, 0x03, 1, D_UINT8, "Update device" },
    { 0x80, 0x04, 1, D_UINT8, "Abort" },
    { 0x80, 0x05, 1, D_UINT8, "Progress" },
    { 0x80, 0x06, 1, D_UINT8, "Error" },
    { 0x80, 0x07, 1, D_UINT16, "Number of parts" },
    { 0x80, 0x08, 1, D_UINT16, "Part size" },
    { 0x80, 0x09, 1, D_UINT16, "Remaining" },
    { 0x81, 0x01, 1, D_UINT8, "Addressing method" },
    { 0x81, 0x02, 1, D_UINT8, "Enable" },
    { 0x81, 0x03, 1, D_UINT32, "FEC delay" },
    { 0x81, 0x04, 1, D_UINT8, "FEC enable" },
    { 0x81, 0x05, 1, D_UINT16, "FEC #columns" },
    { 0x81, 0x06, 1, D_UINT16, "FEC #rows" },
    { 0x81, 0x07, 1, D_UINT32, "IP jitter tolerance" },
    { 0x81, 0x08, 1, D_UINT64, "IP lost after FEC" },
    { 0x81, 0x09, 1, D_UINT64, "IP lost before FEC" },
    { 0x81, 0x0A, 1, D_UINT16, "UDP port" },
    { 0x81, 0x0B, 1, D_UINT16, "IP-to-output delay" },
    { 0x81, 0x0C, 1, D_IP, "Multicast address" },
    { 0x81, 0x0D, 1, D_UINT8, "#TP per IP" },
    { 0x81, 0x0E, 1, D_UINT8, "Status" },
    { 0x81, 0x0F, 1, D_UINT8, "Protocol" },
    { 0x81, 0x10, 1, D_UINT32, "Index" },
    { 0x81, 0x11, 1, D_UINT8, "Output type" },
    { 0x81, 0x12, 1, D_UINT8, "Packet size" },
    { 0x81, 0x13, 1, D_UINT32, "Bitrate" },
    { 0x81, 0x14, 1, D_UINT8, "PCR present" },
    { 0x81, 0x15, 1, D_UINT32, "Rate-change counter" },
    { 0x81, 0x16, 1, D_UINT8, "Rate-estimation mode" },
    { 0x81, 0x17, 1, D_UINT32, "Jitter error counter" },
    { 0x81, 0x18, 1, D_UINT32, "Lock error counter" },
    { 0x81, 0x19, 1, D_UINT32, "Delay factor" },
    { 0x82, 0x01, 1, D_UINT8, "Enable" },
    { 0x82, 0x03, 1, D_UINT8, "FEC enable" },
    { 0x82, 0x04, 1, D_UINT16, "FEC #columns" },
    { 0x82, 0x05, 1, D_UINT16, "FEC #rows" },
    { 0x82, 0x06, 1, D_IP, "IP address" },
    { 0x82, 0x07, 1, D_UINT8, "IP address status" },
    { 0x82, 0x08, 1, D_UINT16, "UDP port" },
    { 0x82, 0x09, 1, D_UINT8, "#TP per IP" },
    { 0x82, 0x0A, 1, D_UINT8, "Protocol" },
    { 0x82, 0x0B, 1, D_UINT32, "Index" },
    { 0x82, 0x0C, 1, D_UINT8, "Input type" },
    { 0x82, 0x0D, 1, D_UINT8, "Time to live" },
    { 0x82, 0x0E, 1, D_IPV6, "IPv6 address" },
    { 0x82, 0x0F, 1, D_UINT8, "FEC dimension" },
    { 0x83, 0x01, 1, D_UINT8, "Packet size" },
    { 0x83, 0x02, 1, D_UINT8, "Physical port" },
    { 0x83, 0x03, 1, D_UINT8, "Status" },
    { 0x83, 0x04, 1, D_UINT32, "TS rate" },
    { 0x83, 0x05, 1, D_UINT64, "Valid count" },
    { 0x83, 0x06, 1, D_UINT64, "Violation count" },
    { 0x83, 0x07, 1, D_UINT8, "Input port select" },
    { 0x83, 0x08, 1, D_UINT8, "Enable" },
    { 0x83, 0x09, 1, D_UINT8, "Polarity" },
    { 0x83, 0x0A, 1, D_UINT8, "Maximum bitrate" },
    { 0x83, 0x0B, 1, D_UINT8, "Current state" },
    { 0x83, 0x0C, 1, D_UINT8, "Detected polarity" },
    { 0x83, 0x0D, 1, D_UINT8, "Detected packet size" },
    { 0x83, 0x0E, 1, D_UINT64, "Bitrate overflow count" },
    { 0x84, 0x01, 1, D_UINT8, "ASI packet size" },
    { 0x84, 0x02, 1, D_UINT8, "Physical port" },
    { 0x84, 0x03, 1, D_UINT32, "TS rate" },
    { 0x85, 0x01, 1, D_UINT8, "Mode" },
    { 0x85, 0x02, 1, D_UINT8, "Erase" },
    { 0x85, 0x03, 1, D_UINT8, "Programming data" },
    { 0x85, 0x04, 1, D_UINT8, "Verify" },
    { 0x86, 0x01, 1, D_UINT16, "Application index" },
    { 0x86, 0x02, 1, D_UINT8, "Command" },
    { 0x86, 0x03, 1, D_UINT8, "Status" },
    { 0x86, 0x04, 1, D_UINT8, "Progress" },
    { 0x86, 0x05, 1, D_UINT16, "Error reason" },
    { 0x86, 0x06, 1, D_UINT32, "Error information" },
    { 0x86, 0x07, 1, D_UINT8, "Transmit" },
    { 0x86, 0x08, 1, D_UINT32, "Buffer available" },
    { 0x90, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x90, 0x01, 1, D_UINT8, "Receive mode" },
    { 0x90, 0x03, 1, D_UINT32, "Frequency" },
    { 0x90, 0x09, 1, D_UINT8, "LNB enable" },
    { 0x90, 0x0A, 1, D_UINT8, "LNB voltage" },
    { 0x90, 0x0B, 1, D_UINT8, "LNB tone" },
    { 0x90, 0x0C, 1, D_UINT8, "DVB-S2 ISI" },
    { 0x90, 0x0D, 1, D_UINT8, "L.3 timestamp insertion" },
    { 0x90, 0x0E, 1, D_UINT8, "Tuner mode" },
    { 0x90, 0x0F, 1, D_UINT8, "Tuner bandwidth" },
    { 0x91, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x91, 0x01, 1, D_UINT8, "Locked" },
    { 0x91, 0x02, 1, D_UINT32, "Frequency" },
    { 0x91, 0x03, 1, D_UINT32, "Bandwidth" },
    { 0x91, 0x04, 1, D_UINT8, "Modulation type" },
    { 0x91, 0x05, 1, D_UINT8, "Code rate" },
    { 0x91, 0x06, 1, D_UINT8, "Roll-off factor" },
    { 0x91, 0x07, 1, D_UINT8, "FEC Frame length" },
    { 0x91, 0x08, 1, D_UINT8, "Pilots" },
    { 0x91, 0x09, 1, D_UINT8, "Spectrum inversion" },
    { 0x91, 0x0A, 1, D_UINT32, "Symbol rate" },
    { 0x91, 0x0B, 1, D_UINT32, "Bitrate" },
    { 0x92, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x92, 0x01, 10, D_INT32, "RF power level" },
    { 0x92, 0x02, 10, D_INT32, "SNR" },
    { 0x92, 0x03, 10, D_INT32, "MER" },
    { 0x92, 0x04, 10, D_INT32, "Es/N0" },
    { 0x92, 0x05, 10, D_INT32, "Eb/N0" },
    { 0x92, 0x06, 10, D_INT32, "Link margin" },
    { 0x92, 0x07, 1, D_UINT64, "BER pre-Viterbi" },
    { 0x92, 0x08, 1, D_UINT64, "BER post-Viterbi" },
    { 0x92, 0x09, 1, D_UINT64, "BER pre-Reed Solomon" },
    { 0x92, 0x0A, 1, D_UINT64, "BER pre-LDPC" },
    { 0x92, 0x0B, 1, D_UINT64, "BER post-LDPC" },
    { 0x92, 0x0C, 1, D_UINT64, "BER post-BCH" },
    { 0x92, 0x0D, 1, D_UINT32, "AGC1" },
    { 0x92, 0x0E, 1, D_UINT32, "AGC2" },
    { 0x93, 0x01, 1, D_UINT8, "Burst" },
    { 0x93, 0x02, 1, D_UINT8, "Send DiSEqC message" },
    { 0x93, 0x03, 1, D_UINT8, "Send DiSEqC message and receive response" },
    { 0x93, 0x04, 1, D_UINT8, "DiSEqC response message length" },
    { 0x93, 0x05, 1, D_UINT8, "DiSEqC response message" },
    { 0x94, 0x00, 1, D_UINT8, "All settings in this category (RO)" },
    { 0x94, 0x01, 1, D_UINT8, "State" },
    { 0x94, 0x02, 1, D_UINT32, "Transport-Stream ID" },
    { 0x94, 0x03, 1, D_UINT32, "Network ID" },
    { 0x94, 0x04, 1, D_UINT32, "Frequency" },
    { 0x94, 0x05, 1, D_UINT32, "Orbital Position" },
    { 0x94, 0x06, 1, D_UINT8, "West east flag" },
    { 0x94, 0x07, 1, D_UINT8, "Polarization" },
    { 0x94, 0x08, 1, D_UINT8, "Roll-off factor" },
    { 0x94, 0x09, 1, D_UINT8, "Modulation system" },
    { 0x94, 0x0A, 1, D_UINT8, "Modulation Type" },
    { 0x94, 0x0B, 1, D_UINT32, "Symbol rate" },
    { 0x94, 0x0C, 1, D_UINT8, "Code rate (FEC Inner)" }
};
const int  g_NumDtmSettingInfoItems = sizeof(g_DtmSettingInfo) 
                                                           / sizeof(g_DtmSettingInfo[0]);

// This table indicates for each DTM-32xx device whether the subtype setting is available
const DtmDeviceSubTypePresence  g_DtmDeviceSubTypePresence[] =
{
    { 3200, 0 },
    { 3237, 1 }
};
const int  g_NumDtmDeviceSubTypePresence = sizeof(g_DtmDeviceSubTypePresence) 
                                                  / sizeof(g_DtmDeviceSubTypePresence[0]);

// This table defines the firmware-upgrade type for each DTM-32xx device
const DtmUpgradeType  g_DtmUpgradeTypes[] =
{
    { 3200, 1 },
    { 3237, 2 },
    { 3224, 3 },
};
const int  g_NumDtmUpgradeTypes = sizeof(g_DtmUpgradeTypes) 
                                                           / sizeof(g_DtmUpgradeTypes[0]);

// This table defines the firmware character encoding as a function of DTM-32xx / version
const DtmFwFileCharEncoding  g_DtmFwFileCharEncoding[] =
{
    { 3200, 0, 0, 0 },              // No ASC128 for DTM-3200 firmware prior to v1.4
    { 3200, 14, 0, 1 },             // ASC128 encoding for DTM-3200 firmware >= v1.4
    { 3237, 0, 0, 1}                // ASC128 encoding for DTM-3237 bootloader >= v0
};
const int  g_NumDtmFwFileCharEncoding = sizeof(g_DtmFwFileCharEncoding) 
                                                     / sizeof(g_DtmFwFileCharEncoding[0]);

//-.-.-.-.-.-.-.-.-.-.-.- Forward Declarations of Helper Functions -.-.-.-.-.-.-.-.-.-.-.-

// High-level helper functions
DtmResult  DtmDoFirmwareUpgrade(int DtmAddr, int FileSize, int UpgradeType, 
                            DtmUpgGetData pGetData, DtmUpgReportProgress pReportProgress);
DtmResult  DtmReboot(int DtmAddr);
DtmResult  DtmSendCommand(int IgnoreReply);
DtmResult  DtmSendCommandWithRetry(int IgnoreReply);
DtmResult  DtmWriteBytesInt(int DtmAddr, int Category, int Setting, int Index, 
                                                               char* pData, int NumBytes);
int  GetFieldSize(int Category, int Setting);
int  IsCategoryWithIndex(int Category);

// Low-level helper functions
void  DtmIntToHex(int Input, int Bytes, char* pHex);
int  DtmByteToHex(unsigned char* pByte, int Len, char* pHex, int InputDataIsMsb);
int  DtmHexToByte(char* pHex, int Len, char* pReturnData, int InputDataIsMsb);
unsigned long long  DtmHexToInt(char* pHex, int Len);
int  DtmGetChecksum(char* pData, int Len, int StartValue);
int  BinToAsc128(unsigned char* pDataIn, int LengthIn, char* pDataOut);

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Internal Variables -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Miscellaneous global variables
static DtmComType  g_ComType = UART;
static DtmPumpFunction  g_pDtmPumpFunction = (DtmPumpFunction)NULL;
static DtmRxFunction  g_pDtmRxFunction = (DtmRxFunction)NULL;
static DtmSleepFunction  g_pDtmSleepFunction = (DtmSleepFunction)NULL;
static DtmTxFunction  g_pDtmTxFunction = (DtmTxFunction)NULL;
static void*  g_pOpaque = NULL;

// Message buffers
static DtmMessage  g_RxMessage;
static DtmMessage  g_TxMessage;

// Buffers used for firmware upgrade
char  g_ByteBuffer[DTM_UPG_SIZE_HEX];
#ifndef USE_DYN_UPG_BUFFER
char  g_PartData[DTM_UPG_UART_PART_SIZE];
char  g_UpgradeBuffer[DTM_UPG_SIZE_HEX];
#endif

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ DTM Handler API Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmHandlerInit -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// DTM handler initialization. Use this function to set the appropriate communication
// type (UART or I2C) and to hook up the platform-specific communication functions
//
void  DtmHandlerInit(void* pOpaque, DtmComType ComType, DtmRxFunction pRxFunction, 
                               DtmTxFunction pTxFunction, DtmSleepFunction pSleepFunction, 
                               DtmPumpFunction pPumpFunction)
{
    g_ComType = ComType;
    g_pDtmRxFunction = pRxFunction;
    g_pDtmPumpFunction = pPumpFunction;
    g_pDtmSleepFunction = pSleepFunction;
    g_pDtmTxFunction = pTxFunction;
    g_pOpaque = pOpaque;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmReadBytes -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Reads bytes from the DTM device and stores these in pData, starting at pData[0].
//
DtmResult  DtmReadBytes(int DtmAddr, int Category, int Setting, int Index, char* pData, 
                                                                             int NumBytes)
{
    DtmResult  Result;

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<0 || pData==NULL 
                                                           || NumBytes<=0 || NumBytes>100)
        return DTM_E_INVALID_PARAMETER;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_READ;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = 0;
    g_TxMessage.m_ConvBinToHex = 0;
    g_TxMessage.m_WriteBytes = 0;
    g_TxMessage.m_pData = 0;
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = NumBytes;
    g_RxMessage.m_ReadBytes = 1;
    g_RxMessage.m_pData = pData;

    Result = DtmSendCommandWithRetry(0);

    if (Result != DTM_OK)
        return Result;

    if (g_RxMessage.m_DataLength == 0)
        return DTM_GOT_NO_DATA;

    if (g_RxMessage.m_DataLength < NumBytes)
    {
        NumBytes = g_RxMessage.m_DataLength;
        Result = DTM_GOT_LESS_DATA;
    }

    if (g_RxMessage.m_DataLength > NumBytes)
        Result = DTM_GOT_MORE_DATA;

    return Result;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmReadInt -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
DtmResult  DtmReadInt(int DtmAddr, int Category, int Setting, int Index, int* pValue)
{
    DtmResult  Result;
    int  FieldSize = GetFieldSize(Category, Setting);

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || pValue==NULL 
                                                           || FieldSize<=0 || FieldSize>4)
        return DTM_E_INVALID_PARAMETER;
    
    // Initialize value to 0
    *pValue = 0;

    // Initialize Rx and Tx message structures
    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_READ;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = 0;
    g_TxMessage.m_ConvBinToHex = 0;
    g_TxMessage.m_WriteBytes = 0;
    g_TxMessage.m_pData = 0;

    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = FieldSize;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = (char *)(pValue);

    // Send DTM command
    Result = DtmSendCommandWithRetry(0);

    // Check result
    if (Result != DTM_OK)
        return Result;

    if (g_RxMessage.m_DataLength == 0)
        return DTM_GOT_NO_DATA;

    if (g_RxMessage.m_DataLength < FieldSize)
    {
        FieldSize = g_RxMessage.m_DataLength;
        Result = DTM_GOT_LESS_DATA;
    }

    if (g_RxMessage.m_DataLength > FieldSize)
        Result = DTM_GOT_MORE_DATA;

    return Result;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmReadInt64 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
DtmResult  DtmReadInt64(int DtmAddr, int Category, int Setting,
                                                    int Index, unsigned long long* pValue)
{
    DtmResult  Result;
    int  FieldSize = GetFieldSize(Category, Setting);

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || pValue==NULL 
                                                           || FieldSize<=0 || FieldSize>8)
        return DTM_E_INVALID_PARAMETER;
    
    // Initialize value to 0
    *pValue = 0;

    // Initialize Rx and Tx message structures
    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_READ;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = 0;
    g_TxMessage.m_ConvBinToHex = 0;
    g_TxMessage.m_WriteBytes = 0;
    g_TxMessage.m_pData = 0;

    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = FieldSize;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = (char *)(pValue);

    // Send DTM command
    Result = DtmSendCommandWithRetry(0);

    // Check result
    if (Result != DTM_OK)
        return Result;

    if (g_RxMessage.m_DataLength == 0)
        return DTM_GOT_NO_DATA;

    if (g_RxMessage.m_DataLength < FieldSize)
    {
        FieldSize = g_RxMessage.m_DataLength;
        Result = DTM_GOT_LESS_DATA;
    }

    if (g_RxMessage.m_DataLength > FieldSize)
        Result = DTM_GOT_MORE_DATA;

    return Result;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmWriteBytes -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Transmits bytes supplied through pData to the DTM device, starting at pData[0].
//
DtmResult  DtmWriteBytes(int DtmAddr, int Category, int Setting, int Index, char* pData, 
                                                                             int NumBytes)
{
    char  RxDataBuffer[300];

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || NumBytes<=0)
        return DTM_E_INVALID_PARAMETER;

    if (g_ComType==UART && NumBytes>DTM_UPG_UART_PART_SIZE*2)
        return DTM_E_INVALID_PARAMETER;

    if (g_ComType==I2C && NumBytes>DTM_UPG_I2C_PART_SIZE)
        return DTM_E_INVALID_PARAMETER;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_WRITE;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = NumBytes;
    g_TxMessage.m_ConvBinToHex = 1;
    g_TxMessage.m_WriteBytes = 1;
    g_TxMessage.m_pData = pData;
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = RxDataBuffer;
    
    // The response data length depends on the category and setting
    if (Category == 0x80 && Setting == 0x02)
        // Firmware upgrade file part: Return message contains file-part sequence number
        g_RxMessage.m_DataLength = 2;
    else if (Category == 0x85 && Setting == 0x03)
        // Firmware upgrade programming data: Return message contains no data
        g_RxMessage.m_DataLength = 0;
    else
        // Otherwise, response message will be a copy of the command
        g_RxMessage.m_DataLength = NumBytes;

    return DtmSendCommandWithRetry(0);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmWriteInt -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
DtmResult  DtmWriteInt(int DtmAddr, int Category, int Setting, int Index, int pValue)
{
    int  FieldSize = GetFieldSize(Category, Setting);
    int  RxDataBuffer;
    int  IgnoreReply = 0;

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || FieldSize<=0 
                                                                           || FieldSize>4)
        return DTM_E_INVALID_PARAMETER;

    if ((FieldSize==1 && pValue>0xFF) || (FieldSize==2 && pValue>0xFFFF))
        return DTM_OUT_OF_RANGE;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_WRITE;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = FieldSize;
    g_TxMessage.m_ConvBinToHex = 1;
    g_TxMessage.m_WriteBytes = 0;
    g_TxMessage.m_pData = (char *)(&pValue);
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = FieldSize;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = (char *)(&RxDataBuffer);

    // Ignore reply for reboot command
    if (Category==0x03 && Setting==0x05)
        IgnoreReply = 1;

    return DtmSendCommandWithRetry(IgnoreReply);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmWriteInt64 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
DtmResult  DtmWriteInt64(int DtmAddr, int Category, int Setting,
                                                     int Index, unsigned long long pValue)
{
    int  FieldSize = GetFieldSize(Category, Setting);
    unsigned long long  RxDataBuffer;

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || FieldSize<=0 
                                                                           || FieldSize>8)
        return DTM_E_INVALID_PARAMETER;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_WRITE;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = FieldSize;
    g_TxMessage.m_ConvBinToHex = 1;
    g_TxMessage.m_WriteBytes = 0;
    g_TxMessage.m_pData = (char *)(&pValue);
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = FieldSize;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = (char *)(&RxDataBuffer);

    return DtmSendCommandWithRetry(0);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmGetUpgradeType -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Get the upgrade type.
// DTM-3200 uses firmware upgrade method 1, whereas DTM-3237 uses method 2.
//
DtmResult  DtmGetUpgradeType(int DtmAddr, int* pUpgradeType)
{
    DtmResult  Result;
    int  i, DeviceType;

    // Get device type
    Result = DtmReadInt(DtmAddr, 0x01, 0x05, 0x00, &DeviceType);
    if (Result != DTM_OK)
        return Result;

    // Switch to firmware upgrade mode for DTM type 2 devices
    for (i=0; i<g_NumDtmUpgradeTypes; i++)
    {
        if (g_DtmUpgradeTypes[i].m_Type == DeviceType)
        {
            *pUpgradeType = g_DtmUpgradeTypes[i].m_UpgradeType;
            return DTM_OK;
        }
    }

    return DTM_E_NOT_SUPPORTED;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmFirmwareUpgrade -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Upgrade the firmware of a DTM device
//
// Parameters:
//  DtmAddr         Address of the targeted DTM-32xx
//  FileSize        Upgrade filesize in bytes
//  pGetData        Callback function to retrieve upgrade data from firmware upgrade file
//  pReportProgress Optional callback function to pass Firmware Upgrade progress
//
// Returns:
//  DTM_OK - when everything was handled correctly
//  DTM_E_TXFUNC_NULL_PNTR - when the tx callback pointer was not set
//  DTM_E_RXFUNC_NULL_PNTR - when the rx callback pointer was not set
//  DTM_E_UPG_NULL_PNTR - when the callback pointer to retrieve file data was not set
//  DTM_E_TX_FAILED - when the tx callback returns this value
//  DTM_E_RX_FAILED - when the rx callback returns this value
//  DTM_E_INCOMPLETE_MSG - when an incomplete message was received
//  DTM_E_INCORRECT_CHECKSUM - when a message with incorrect checksum was received
//  DTM_E_OUT_OF_MEM - when a dynamic upgrade buffer is used and not enough memory is 
//                     available
//  DTM_E_ERROR_REPLY - when the DTM-32xx responded with an 'E' reply
//  DTM_E_BUSY_REPLY - when the DTM-32xx responded with an 'B' reply
//
DtmResult  DtmFirmwareUpgrade(int DtmAddr, int FileSize, DtmUpgGetData pGetData, 
                                                     DtmUpgReportProgress pReportProgress)
{
    DtmResult  Result;
    int  UpgradeType;

    // Check input parameters
    if (DtmAddr<=0 || DtmAddr>0xFF || FileSize<=0 || pGetData==NULL)
        return DTM_E_INVALID_PARAMETER;

    // Get upgrade type
    Result = DtmGetUpgradeType(DtmAddr, &UpgradeType);
    if (Result != DTM_OK)
        return Result;

    return DtmDoFirmwareUpgrade(DtmAddr, FileSize, UpgradeType, pGetData, 
                                                                         pReportProgress);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmWaitForReady -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// This function reads the Busy setting (category 0x01, setting 0x0C) untill the device
// is ready or an error occured
//
DtmResult  DtmWaitForReady(int DtmAddr, int IgnoreErrors)
{
    int  Busy;
    do
    {
        DtmResult  Result = DtmReadInt(DtmAddr, 1, 0x0C, 0x00, &Busy);
        if (Result!=DTM_OK && !IgnoreErrors)
            return Result;
        if (Result!=DTM_OK && IgnoreErrors)
            // Data might be corrupted when errors are received, so set busy to safe state
            Busy = 1;
        
        // Sleep if the device is busy
        if (Busy != 0)
            g_pDtmSleepFunction(g_pOpaque, 1);
    }
    while (Busy != 0);

    return DTM_OK;
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Internal helper functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmReadBytesInt -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Reads bytes from the DTM device and stores these in pRxData, starting at pRxData[0].
//
DtmResult  DtmReadBytesInt(int DtmAddr, int Category, int Setting, int Index, 
                             char* pTxData, int TxNumBytes, char* pRxData, int RxNumBytes)
{
    DtmResult  Result;

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<0 || pTxData==NULL 
                                       || pRxData==NULL || TxNumBytes<=0 || TxNumBytes>100 
                                       || RxNumBytes<=0 || RxNumBytes>100)
        return DTM_E_INVALID_PARAMETER;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_READ;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = TxNumBytes;
    g_TxMessage.m_ConvBinToHex = 1;
    g_TxMessage.m_WriteBytes = 1;
    g_TxMessage.m_pData = pTxData;
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_DataLength = RxNumBytes;
    g_RxMessage.m_ReadBytes = 1;
    g_RxMessage.m_pData = pRxData;

    Result = DtmSendCommandWithRetry(0);

    if (Result != DTM_OK)
        return Result;

    if (g_RxMessage.m_DataLength == 0)
        return DTM_GOT_NO_DATA;

    if (g_RxMessage.m_DataLength < RxNumBytes)
    {
        RxNumBytes = g_RxMessage.m_DataLength;
        Result = DTM_GOT_LESS_DATA;
    }

    if (g_RxMessage.m_DataLength > RxNumBytes)
        Result = DTM_GOT_MORE_DATA;

    return Result;
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+ Protocol Implementation Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmFirmwareUpgradeType1 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Firmware upgrade implementation
//
DtmResult  DtmDoFirmwareUpgrade(int DtmAddr, int FileSize, int UpgradeType, 
                             DtmUpgGetData pGetData, DtmUpgReportProgress pReportProgress)
{
    int  FirmwareVersion;
    int  BootloaderVersion;
    int  Mode;
    int  PartLength;
    int  NumOfParts;
    int  CurrentPartLength;
    int  Temp, i, Error;
    char*  pData;                   
    char*  pPartData;               // PartData buffer will contain binary file-part data
                                    // and is used for ASCII conversion
    DtmResult  Result;
    char  Progress = 0;
    DtmUpgPhase  Phase = INIT;
    int  DeviceType;
    int  UseAsc128Encoding;
    int  FirmwareValid;

    pData = (char*)NULL;
    pPartData = (char*)NULL;

    if (g_pDtmTxFunction == NULL)
        return DTM_E_TXFUNC_NULL_PNTR;
    if (g_pDtmRxFunction == NULL)
        return DTM_E_RXFUNC_NULL_PNTR;
    if (pGetData == NULL)
        return DTM_E_UPG_NULL_PNTR;
    if (UpgradeType!=1 && UpgradeType!=2 && UpgradeType!=3)
        return DTM_E_INVALID_PARAMETER;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Get DTM device type
    Result = DtmReadInt(DtmAddr, 0x01, 0x05, 0x00, &DeviceType);
    if (Result != DTM_OK)
        return Result;

    // Find out if this firmware version expects ASC128 encoding
    if (UpgradeType == 1)
    {
        // Read the DTM firmware version
        Result = DtmReadInt(DtmAddr, 0x01, 0x03, 0x00, &FirmwareVersion);
        if (Result != DTM_OK)
            return Result;

        for (i=0; i<g_NumDtmFwFileCharEncoding; i++)
            if (g_DtmFwFileCharEncoding[i].m_Type==DeviceType 
                         && FirmwareVersion>=g_DtmFwFileCharEncoding[i].m_FirmwareVersion)
                UseAsc128Encoding = g_DtmFwFileCharEncoding[i].m_UseAsc128Encoding;
    } 
    else if (UpgradeType == 2)
    {
        // Read the DTM bootloader version
        Result = DtmReadInt(DtmAddr, 0x01, 0x07, 0x00, &BootloaderVersion);
        if (Result != DTM_OK)
            return Result;

        // Find out if this bootloader version expects ASC128 encoding
        for (i=0; i<g_NumDtmFwFileCharEncoding; i++)
            if (g_DtmFwFileCharEncoding[i].m_Type==DeviceType 
                     && BootloaderVersion>=g_DtmFwFileCharEncoding[i].m_BootloaderVersion)
                UseAsc128Encoding = g_DtmFwFileCharEncoding[i].m_UseAsc128Encoding;
    }
    else if (UpgradeType == 3)
    {
        UseAsc128Encoding = 1;
    }

    // Read mode and verify that the current mode is firmware upgrade
    if (UpgradeType == 2)
    {
        Result = DtmReadInt(DtmAddr, 0x85, 0x01, 0x00, &Mode); 
        if (Result != DTM_OK)
            return Result;

        if (Mode != 1)
            return DTM_E_INVALID_MODE;
    }

    // Erase current firmware
    if (UpgradeType == 2)
    {
        Result = DtmWriteInt(DtmAddr, 0x85, 0x02, 0x00, 1); 
        if (Result != DTM_OK)
            return Result;

        // Wait for the device to be ready: Read the busy flag. Ignore communication 
        // errors, since the device is not responsive during an erase action
        Result = DtmWaitForReady(DtmAddr, 1); 
        if (Result != DTM_OK)
            return Result;
    }
    
    Progress += DTM_UPG_INIT_PROGRESS_INCR;

    if (g_ComType == UART)
        PartLength = DTM_UPG_UART_PART_SIZE;
    else
        PartLength = DTM_UPG_I2C_PART_SIZE;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

#ifdef USE_DYN_UPG_BUFFER
    if (g_ComType == UART)
    {
        if (UseAsc128Encoding)
            pData = (char*)malloc(DTM_UPG_SIZE_ASC);
        else
            pData = (char*)malloc(DTM_UPG_SIZE_HEX);

        if (pData == NULL)
            return DTM_E_OUT_OF_MEM;

        pPartData = (char*)malloc(PartLength);

        if (pPartData == NULL)
        {
            free(pData);
            return DTM_E_OUT_OF_MEM;
        }
    } else {
        pData = (char*)malloc(DTM_UPG_SIZE_I2C);
        pPartData = NULL;
        if (pData == NULL)
            return DTM_E_OUT_OF_MEM;
    }
#else
    pData = g_UpgradeBuffer;
    pPartData = g_PartData;
#endif

    // Calculate the number of file-parts to send to the DTM
    NumOfParts = FileSize / PartLength + (FileSize%PartLength?1:0);

    // Abort any upgrade still pending
    if (UpgradeType == 1)
    {
        Result = DtmWriteInt(DtmAddr, 0x80, 0x04, 0x00, 1);
        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }
    }
    if (UpgradeType == 3)
    {
        // Write 'start upgrade' command
        Result = DtmWriteInt(DtmAddr, 0x86, 0x02, 0x00, 0);
        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }
    }

    Progress += DTM_UPG_INIT_PROGRESS_INCR;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Write the number of file parts to the DTM
    if (UpgradeType == 1)
    {
        Result = DtmWriteInt(DtmAddr, 0x80, 0x07, 0x00, NumOfParts);

        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }
    }
    Progress += DTM_UPG_INIT_PROGRESS_INCR;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Write the length of each (except the last) file-part to the DTM
    if (UpgradeType == 1)
    {
        Result = DtmWriteInt(DtmAddr, 0x80, 0x08, 0x00, PartLength);
        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }
    }
    Progress += DTM_UPG_INIT_PROGRESS_INCR;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Write the length of the last file part to the DTM
    if (UpgradeType == 1)
    {
        Temp = FileSize - (NumOfParts-1)*PartLength;
        Result = DtmWriteInt(DtmAddr, 0x80, 0x09, 0x00, Temp);
        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }
    }
    Progress += DTM_UPG_INIT_PROGRESS_INCR;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    Phase = UPLOADING;
    Progress = 0;

    for (i=0; i<NumOfParts; i++)
    {
        int  Index = i+1;
        int  NumBytes;
        Temp = 0;

        if (g_pDtmPumpFunction != NULL)
            g_pDtmPumpFunction(g_pOpaque);

        if ((FileSize-i*PartLength) < PartLength)
            CurrentPartLength = FileSize - i*PartLength;
        else
            CurrentPartLength = PartLength;

        // Call the function pointer to retrieve the upgrade data
        if (g_ComType == UART)
            // For the UART write it to the PartData buffer to be able to convert it to
            // ASCII later on
            Result = pGetData(g_pOpaque, pPartData, CurrentPartLength);
        else
            // For I2C the data will be send as it is, so we can directly write it to
            // the Data buffer=
            Result = pGetData(g_pOpaque, pData, CurrentPartLength);

        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }

        if (g_pDtmPumpFunction != NULL)
            g_pDtmPumpFunction(g_pOpaque);

        // Convert binary data to ASCII if the UART is used
        if (g_ComType == UART)
        {
            if (UseAsc128Encoding)
                // Asc128 encoding is used
                NumBytes = BinToAsc128((unsigned char *)(pPartData), CurrentPartLength, 
                                                                                   pData);
            else
                // Hex encoding is used
                NumBytes = DtmByteToHex((unsigned char *)(pPartData), CurrentPartLength, 
                                                                                pData, 1);
        } else
            NumBytes = CurrentPartLength;

        if (g_pDtmPumpFunction != NULL)
            g_pDtmPumpFunction(g_pOpaque);
        
        // Transmit the data
        if (UpgradeType == 1)
        {
            g_pDtmSleepFunction(g_pOpaque, 5);
            Result = DtmWriteBytesInt(DtmAddr, 0x80, 0x02, Index, pData, NumBytes);
        }
        else if (UpgradeType == 2)
            Result = DtmWriteBytesInt(DtmAddr, 0x85, 0x03, Index, pData, NumBytes);
        else if (UpgradeType == 3)
            Result = DtmWriteBytesInt(DtmAddr, 0x86, 0x07, Index, pData, NumBytes);
        
        if (Result != DTM_OK)
        {
#ifdef USE_DYN_UPG_BUFFER
            free(pData);
            if (pPartData != NULL)
                free(pPartData);
#endif
            return Result;
        }

        Progress = (100*i) / NumOfParts;

        if (pReportProgress != NULL)
            pReportProgress(g_pOpaque, Phase, Progress);

        if (g_pDtmPumpFunction != NULL)
            g_pDtmPumpFunction(g_pOpaque);
    }

#ifdef USE_DYN_UPG_BUFFER
    if (pData != NULL)
        free(pData);
    if (pPartData != NULL)
        free(pPartData);
#endif

    // Update device
    if (UpgradeType == 1)
    {
        Result = DtmWriteInt(DtmAddr, 0x80, 0x03, 0x00, 1); 
        if (Result != DTM_OK)
            return Result;
    }
    else if (UpgradeType == 3)
    {
        // Send 'flush image' command
        Result = DtmWriteInt(DtmAddr, 0x86, 0x02, 0x00, 1);
        if (Result != DTM_OK)
            return Result;
    }
    
    Progress = 0;
    Phase = PROGRAMMING;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Read programming progress
    if (UpgradeType == 1)
    {
        do {
            // Check for errors
            Error = 0;
            Result = DtmReadInt(DtmAddr, 0x80, 0x06, 0x00, &Error); 
            if (Error)
                return DTM_E_UPG_FAILED;

            // Read progress
            Result = DtmReadInt(DtmAddr, 0x80, 0x05, 0x00, &Temp);
            if (Result != DTM_OK)
            {
                g_pDtmSleepFunction(g_pOpaque, 1);
                continue;
            }

            Progress = (char)(Temp & 0xFF);

            if (pReportProgress != NULL)
                pReportProgress(g_pOpaque, Phase, Progress);
            if (g_pDtmPumpFunction != NULL)
                g_pDtmPumpFunction(g_pOpaque);
        } while(Progress < 100);
    }
    else if (UpgradeType == 3)
    {
        int  Status;
        do {
            // Don't go too fast
            g_pDtmSleepFunction(g_pOpaque, 300);

            // Check for errors
            Error = 0;
            Result = DtmReadInt(DtmAddr, 0x86, 0x03, 0x00, &Status);
            if (Status!=1 && Status!=2)
            {
                int  ErrorReason, ErrorInfo;
                DtmReadInt(DtmAddr, 0x86, 0x05, 0x00, &ErrorReason);
                DtmReadInt(DtmAddr, 0x86, 0x06, 0x00, &ErrorInfo);
                return DTM_E_UPG_FAILED;
            }

            // Read progress
            Result = DtmReadInt(DtmAddr, 0x86, 0x04, 0x00, &Temp);
            if (Result != DTM_OK)
            {
                g_pDtmSleepFunction(g_pOpaque, 1);
                continue;
            }

            Progress = (char)(Temp & 0xFF);

            if (pReportProgress != NULL)
                pReportProgress(g_pOpaque, Phase, Progress);
            if (g_pDtmPumpFunction != NULL)
                g_pDtmPumpFunction(g_pOpaque);
        } while (Status != 2);
    }

    // Verify uploaded firmware
    if (UpgradeType == 2)
    {
        Result = DtmReadInt(DtmAddr, 0x85, 0x04, 0x00, &FirmwareValid);
        if (Result != DTM_OK)
            return Result;

        if (FirmwareValid != 1)
            return DTM_E_UPG_VERIFY_FAILED;
    }
    else if (UpgradeType == 3)
    {
        int  Status;

        // Send 'Start verification' command
        Result = DtmWriteInt(DtmAddr, 0x86, 0x02, 0x00, 2);
        if (Result != DTM_OK)
            return Result;

        do {
            // Don't go too fast
            g_pDtmSleepFunction(g_pOpaque, 300);

            // Check for errors
            Error = 0;
            Result = DtmReadInt(DtmAddr, 0x86, 0x03, 0x00, &Status); 
            if (Status!=3 && Status!=4)
                return DTM_E_UPG_VERIFY_FAILED;

            if (g_pDtmPumpFunction != NULL)
                g_pDtmPumpFunction(g_pOpaque);
        } while (Status != 4);
    }
    
    // Activate new image
    if (UpgradeType == 3)
    {
        int  Status;

        // Send 'Activate upgrade' command
        Result = DtmWriteInt(DtmAddr, 0x86, 0x02, 0x00, 3);
        if (Result != DTM_OK)
            return Result;

        do {
            // Don't go too fast
            g_pDtmSleepFunction(g_pOpaque, 300);

            // Check for errors
            Error = 0;
            Result = DtmReadInt(DtmAddr, 0x86, 0x03, 0x00, &Status); 
            if (Status!=5 && Status!=6)
                return DTM_E_UPG_FAILED;

            if (g_pDtmPumpFunction != NULL)
                g_pDtmPumpFunction(g_pOpaque);
        } while (Status != 6);
    }

    Phase = FINISHED;
    Progress = 100;

    if (pReportProgress != NULL)
        pReportProgress(g_pOpaque, Phase, Progress);
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // The automatic reboot was removed starting from DTM-3200 firmware v1.4
    if ((DeviceType==3200 && FirmwareVersion>=14) || UpgradeType==3)
        DtmReboot(DtmAddr);

    return Result;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmReboot -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Send a reboot command to the DTM-32xx
//
DtmResult  DtmReboot(int DtmAddr)
{
    int  Category = 0x03;
    int  Setting = 0x05;
    
    return DtmWriteInt(DtmAddr, Category, Setting, 0, 1);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmSendCommand -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// The DtmSendCommand function takes data from the g_TxMessage structure, parses it
// to I2C or UART byte data, then calls the I2C/UART Tx callback to transmit the message.
// When coming back from the Tx Callback the Rx Callback is called to get the answer
// to the send message, the data returned by the Rx Callback is parsed into the
// g_RxMessage structure.
//
// Returns:
//  DTM_OK                   - Everything was handled correctly
//  DTM_E_TXFUNC_NULL_PNTR   - The Tx callback pointer was not set
//  DTM_E_RXFUNC_NULL_PNTR   - The Rx callback pointer was not set
//  DTM_E_TX_FAILED          - The Tx callback returns this value
//  DTM_E_RX_FAILED          - The Rx callback returns this value
//  DTM_E_INCOMPLETE_MSG     - An incomplete message was received
//  DTM_E_INCORRECT_CHECKSUM - A message with incorrect checksum was received
//  DTM_REPLY_IGNORED        - Everything was handled correctly but the user requested
//                             no reply (IgnoreReply parameter)
DtmResult  DtmSendCommand(int IgnoreReply)
{
    DtmResult  Result;

    int  TxLength, RxLength;

    int  BufferIndex, i;
    int  RxNumBytes;
    char  Temp;
    int  MaxMsgTimeout = -1;
    int  CategoryIdx;

    if (g_pDtmTxFunction == NULL)
        return DTM_E_TXFUNC_NULL_PNTR;
    if (g_pDtmRxFunction == NULL)
        return DTM_E_RXFUNC_NULL_PNTR;
    if (g_pDtmSleepFunction == NULL)
        return DTM_E_RXFUNC_NULL_PNTR;
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Get worst-case response time (max message tmeout)
    for (CategoryIdx=0; CategoryIdx<g_NumDtmCategoryInfoItems; CategoryIdx++)
    {
        if (g_DtmCategoryInfo[CategoryIdx].m_Category == g_TxMessage.m_Category)
        {
            MaxMsgTimeout = g_DtmCategoryInfo[CategoryIdx].m_MaxMsgTimeOut;
            break;
        }
    }
    if (MaxMsgTimeout == -1)
        return DTM_E_INVALID_PARAMETER;

    if (IgnoreReply == 0)
    {
        // Determine receive buffer size
        RxLength = g_RxMessage.m_DataLength;
        RxLength += 4; // addr/cat/set/chk
        if (IsCategoryWithIndex(g_TxMessage.m_Category))
            RxLength += 2;
        if (g_ComType == UART)
        {
            RxLength = RxLength*2 + 2;
            RxLength++; // also for receive buffer used later on
        }
        RxLength++; //RW byte
    }

    // Determine transmit buffer size
    TxLength = 4;                       // address + category + setting + checksum
    if (IsCategoryWithIndex(g_TxMessage.m_Category))  
        TxLength += 2;                  // add 2 bytes when an index is required
    if (g_ComType == UART)
    {
        TxLength = TxLength*2 + 2;      // Add encoding overhead and stx/etx characters
        TxLength++;                     // Add a byte for 0 termination
    }
    TxLength++;                         // Add R/W byte/character

    if (g_TxMessage.m_DataLength > 0)
    {
        if (g_ComType == UART)
            TxLength += g_TxMessage.m_DataLength*2;
        else
            TxLength += g_TxMessage.m_DataLength;
    }

    // Fill transmit buffer
    BufferIndex = 0;
    if (g_ComType == UART)
    {
        g_ByteBuffer[BufferIndex++] = 0x02; //stx
        BufferIndex += DtmByteToHex((unsigned char*)(&g_TxMessage.m_Address), 1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
        BufferIndex += DtmByteToHex((unsigned char*)(&g_TxMessage.m_Category), 1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
        BufferIndex += DtmByteToHex((unsigned char*)(&g_TxMessage.m_Setting), 1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
        g_ByteBuffer[BufferIndex++] = 
                               (g_TxMessage.m_ReadWrite == DTM_FIELD_RW_READ) ? 'r' : 'w';
        if (IsCategoryWithIndex(g_TxMessage.m_Category))
        {
            Temp = (char)(g_TxMessage.m_Index >> 8 & 0xFF);
            BufferIndex += DtmByteToHex((unsigned char*)(&Temp), 1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
            Temp = (char)(g_TxMessage.m_Index & 0xFF);
            BufferIndex += DtmByteToHex((unsigned char*)(&Temp), 1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
        }
        if (g_TxMessage.m_DataLength > 0)
        {
            if (g_TxMessage.m_ConvBinToHex)
                // Convert binary data to ASCII hex
                BufferIndex += DtmByteToHex((unsigned char*)(g_TxMessage.m_pData), 
                                     g_TxMessage.m_DataLength, &g_ByteBuffer[BufferIndex], 
                                     g_TxMessage.m_WriteBytes);
            else
                // Memcpy input data to TX buffer
                for (i=0; i<g_TxMessage.m_DataLength; i++)
                    g_ByteBuffer[BufferIndex++] = *(g_TxMessage.m_pData + i);
        }
        g_TxMessage.m_Checksum = DtmGetChecksum(&g_ByteBuffer[1], (BufferIndex-1),0);
        BufferIndex += DtmByteToHex((unsigned char*)(&g_TxMessage.m_Checksum),1, 
                                                           &g_ByteBuffer[BufferIndex], 0);
        g_ByteBuffer[BufferIndex] = 0x03; //etx
        g_ByteBuffer[BufferIndex + 1] = '\0';
    } else {
        g_ByteBuffer[BufferIndex++] = g_TxMessage.m_Address;
        g_ByteBuffer[BufferIndex++] = g_TxMessage.m_Category;
        g_ByteBuffer[BufferIndex++] = g_TxMessage.m_Setting;
        g_ByteBuffer[BufferIndex++] = (g_TxMessage.m_ReadWrite == 0)?0:1;
        if (IsCategoryWithIndex(g_TxMessage.m_Category))
        {
            g_ByteBuffer[BufferIndex++] = (char)(g_TxMessage.m_Index>>8 & 0xFF);
            g_ByteBuffer[BufferIndex++] = (char)(g_TxMessage.m_Index & 0xFF);
        }
        if (g_TxMessage.m_DataLength > 0)
        {
            if (g_TxMessage.m_WriteBytes)
            {
                for (i=0; i<g_TxMessage.m_DataLength; i++)
                    g_ByteBuffer[BufferIndex++] = g_TxMessage.m_pData[i];
            } else {
                for (i=g_TxMessage.m_DataLength-1; i>=0; i--)
                    g_ByteBuffer[BufferIndex++] = g_TxMessage.m_pData[i];
            }
        }
        g_TxMessage.m_Checksum = DtmGetChecksum(g_ByteBuffer, BufferIndex, 0);
        g_ByteBuffer[BufferIndex] = g_TxMessage.m_Checksum;
    }
    
    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    // Call callback to transmit the data
    if (g_ComType == UART)
    {
        Result = g_pDtmTxFunction(g_pOpaque, g_TxMessage.m_Address, g_ByteBuffer, 
                                                                           BufferIndex+1);
        if (Result != DTM_OK)
            return Result;
    } else {
        Result = g_pDtmTxFunction(g_pOpaque, g_TxMessage.m_Address, &g_ByteBuffer[1], 
                                                                             BufferIndex);
        if (Result != DTM_OK)
            return Result;
    }

    if (g_pDtmPumpFunction != NULL)
        g_pDtmPumpFunction(g_pOpaque);

    if (IgnoreReply == 0)
    {
        // Allocate receive buffer
        g_ByteBuffer[RxLength - 1] = '\0';
        BufferIndex = 1;

        // Call callback to receive data
        if (g_ComType == UART)
        {
            const int  RwIdx = 7;
            int  IsErr;
            Result = g_pDtmRxFunction(g_pOpaque, g_TxMessage.m_Address, 
                           (char*)(g_ByteBuffer), RxLength-1, &RxNumBytes, MaxMsgTimeout);
            // Catch protocol errors (data is removed from message, so result is shorter)
            IsErr = (g_ByteBuffer[RwIdx]=='e' || g_ByteBuffer[RwIdx]=='E') ? 1 : 0;
            if (Result==DTM_E_TIME_OUT && IsErr && (RxNumBytes==11 || RxNumBytes==15 
                                                                       || RxNumBytes==19))
                return DTM_E_ERROR_REPLY;
            // Return all other interface related errors as is
            if (Result != DTM_OK)
                return Result;

            // Parse received message
            if (g_ByteBuffer[0] != 0x02)
                return DTM_E_INCOMPLETE_MSG;

            i = 0;
            while (g_ByteBuffer[i] != 0x03)
            {
                i++;
                if (i == RxLength)
                    return DTM_E_INCOMPLETE_MSG;
            }

            // Check checksum
            DtmHexToByte(&g_ByteBuffer[i-2], 2, (char*)(&g_RxMessage.m_Checksum), 0);
            if (g_RxMessage.m_Checksum != DtmGetChecksum(&g_ByteBuffer[1], 
                                                                         RxNumBytes-4, 0))
                return DTM_E_INCORRECT_CHECKSUM;

            DtmHexToByte(&g_ByteBuffer[BufferIndex], 2, (char*)(&g_RxMessage.m_Address), 
                                                                                       0);
            BufferIndex += 2;
            DtmHexToByte(&g_ByteBuffer[BufferIndex], 2, (char*)(&g_RxMessage.m_Category),
                                                                                       0);
            BufferIndex += 2;
            DtmHexToByte(&g_ByteBuffer[BufferIndex], 2, (char*)(&g_RxMessage.m_Setting),
                                                                                       0);
            BufferIndex += 2;

            if (g_RxMessage.m_Category!=g_TxMessage.m_Category ||
                                             g_RxMessage.m_Setting!=g_TxMessage.m_Setting)
                return DTM_E_INCORRECT_REPLY;

            switch (g_ByteBuffer[BufferIndex])
            {
                case 'e':
                case 'E': g_RxMessage.m_ReadWrite = 'E'; break;
                case 'b':
                case 'B': g_RxMessage.m_ReadWrite = 'B'; break;
                case 'w':
                case 'W': g_RxMessage.m_ReadWrite = DTM_FIELD_RW_WRITE; break;
                case 'r':
                case 'R': g_RxMessage.m_ReadWrite = DTM_FIELD_RW_READ; break;
                default:
                    g_RxMessage.m_ReadWrite = 0x00;
            }
            BufferIndex++;
            if (IsCategoryWithIndex(g_RxMessage.m_Category))
            {
                g_RxMessage.m_Index = (int)(DtmHexToInt(&g_ByteBuffer[BufferIndex], 4));
                BufferIndex += 4;
            }

            // Data
            g_RxMessage.m_DataLength = ((i-BufferIndex)-2); // ASCII length

            // Copy data to receive buffer
            g_RxMessage.m_DataLength = DtmHexToByte(&g_ByteBuffer[BufferIndex],
                                            g_RxMessage.m_DataLength, g_RxMessage.m_pData, 
                                            g_RxMessage.m_ReadBytes);
        } else {
            // Poll for the response message
            int  TimeOut = 0;
            const int  RwIdx = 3;
            int  IsErr;
            g_ByteBuffer[0] = (unsigned char)(g_TxMessage.m_Address);   // Address is not
                                                                        // received
            
            g_RxMessage.m_Address = g_TxMessage.m_Address;
            while (TimeOut < MaxMsgTimeout)
            {
                Result = g_pDtmRxFunction(g_pOpaque, g_TxMessage.m_Address, 
                                                     &g_ByteBuffer[1], 1, &RxNumBytes, 0);
                if (Result != DTM_OK)
                    return Result;

                // Check if the response message is ready
                if (g_ByteBuffer[1] != 0x00)
                    break;

                // Don't poll too fast
                g_pDtmSleepFunction(g_pOpaque, 1);
                TimeOut++;
                if (TimeOut == MaxMsgTimeout)
                    return DTM_E_TIME_OUT;
            }

            // Response message is ready, retrieve the complete message
            Result = g_pDtmRxFunction(g_pOpaque, g_TxMessage.m_Address, &g_ByteBuffer[1], 
                                                              RxLength-1, &RxNumBytes, 0);
            // Catch protocol errors (data is removed from message, so result is shorter)
            IsErr = (g_ByteBuffer[RwIdx]=='e' || g_ByteBuffer[RwIdx]=='E') ? 1 : 0;
            if (IsErr)
                return DTM_E_ERROR_REPLY;
            // Return all other interface related errors as is
            if (Result != DTM_OK)
                return Result;

            // Parse received message
            g_RxMessage.m_Category = (unsigned char)(g_ByteBuffer[BufferIndex++]);
            g_RxMessage.m_Setting = (unsigned char)(g_ByteBuffer[BufferIndex++]);
            g_RxMessage.m_ReadWrite = (unsigned char)(g_ByteBuffer[BufferIndex++]);
            if (IsCategoryWithIndex(g_RxMessage.m_Category))
            {
                g_RxMessage.m_Index = ((unsigned char)(g_ByteBuffer[BufferIndex]) << 8)
                                         + (unsigned char)(g_ByteBuffer[BufferIndex + 1]);
                BufferIndex += 2;
            }

            // Validate checksum to find out if the expected number of bytes are received
            i = RxLength - 1;
            g_RxMessage.m_Checksum = (unsigned char)(g_ByteBuffer[i]);
            if (g_RxMessage.m_Checksum != DtmGetChecksum(g_ByteBuffer, i, 0))
            {
                // Checksum isn't correct, try to find out the message length
                // Find first non-zero byte starting from end working backwards
                while(g_ByteBuffer[i] == 0x0)
                {
                    i--;
                    if (i < BufferIndex)
                        return DTM_E_INCOMPLETE_MSG;
                }

                // Assume i is checksum location
                g_RxMessage.m_Checksum = (unsigned char)(g_ByteBuffer[i]);
                if (g_RxMessage.m_Checksum != DtmGetChecksum(g_ByteBuffer, i, 0))
                    return DTM_E_INCORRECT_CHECKSUM;
            }

            g_RxMessage.m_DataLength = i - BufferIndex;

            // Copy data to receive buffer
            if (g_RxMessage.m_ReadBytes)
            {
                for (i=0; i<g_RxMessage.m_DataLength; i++)
                    g_RxMessage.m_pData[i] = g_ByteBuffer[BufferIndex++];
            } else {
                for (i=g_RxMessage.m_DataLength-1; i>=0; i--)
                    g_RxMessage.m_pData[i] = g_ByteBuffer[BufferIndex++];
            }
        }

        if (g_pDtmPumpFunction != NULL)
            g_pDtmPumpFunction(g_pOpaque);

        if (g_RxMessage.m_ReadWrite=='E' || g_RxMessage.m_ReadWrite=='e')
            return DTM_E_ERROR_REPLY;

        if (g_RxMessage.m_ReadWrite=='B' || g_RxMessage.m_ReadWrite=='b')
            return DTM_E_BUSY_REPLY;

        return DTM_OK;
    }
    return DTM_REPLY_IGNORED;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmSendCommandWithRetry -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
DtmResult  DtmSendCommandWithRetry(int IgnoreReply)
{
    DtmResult  Result;
    int  RetryCount;

    for (RetryCount=0; RetryCount<3; RetryCount++)
    {
        Result = DtmSendCommand(IgnoreReply);

        // Stop retrying when invalid parameters are given or everything went ok
        if (Result==DTM_OK || Result==DTM_E_ERROR_REPLY || Result==DTM_E_BUSY_REPLY
                                                                     || DTM_REPLY_IGNORED)
            break;
        
        // Flush UART receive buffer when a protocol error occurred
        if (Result!=DTM_OK && g_ComType==UART)
        {
            char  Temp[100];
            int  NumBytes;
            g_pDtmRxFunction(g_pOpaque, g_TxMessage.m_Address, Temp, 100, &NumBytes, 100);
        }
    }

    return Result;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmWriteBytesInt -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Internal helper function for writing firmware data
// Difference with DtmWriteBytes is m_ConvBinToHex, which is set to 0
//
DtmResult  DtmWriteBytesInt(int DtmAddr, int Category, int Setting, int Index, 
                                                                char* pData, int NumBytes)
{
    char  RxDataBuffer[100];

    if (DtmAddr<=0 || DtmAddr>0xFF || Category<=0 || Setting<=0 || NumBytes<=0)
        return DTM_E_INVALID_PARAMETER;

    if (g_ComType==UART && NumBytes>DTM_UPG_UART_PART_SIZE*2)
        return DTM_E_INVALID_PARAMETER;

    if (g_ComType==I2C && NumBytes>DTM_UPG_I2C_PART_SIZE)
        return DTM_E_INVALID_PARAMETER;

    g_TxMessage.m_Address = DtmAddr;
    g_TxMessage.m_Category = Category;
    g_TxMessage.m_Setting = Setting;
    g_TxMessage.m_ReadWrite = DTM_FIELD_RW_WRITE;
    g_TxMessage.m_Index = Index;
    g_TxMessage.m_Checksum = 0;
    g_TxMessage.m_DataLength = NumBytes;
    g_TxMessage.m_ConvBinToHex = 0;
    g_TxMessage.m_WriteBytes = 1;
    g_TxMessage.m_pData = pData;
    
    g_RxMessage.m_Address = 0;
    g_RxMessage.m_Category = 0;
    g_RxMessage.m_Setting = 0;
    g_RxMessage.m_ReadWrite = 0;
    g_RxMessage.m_Index = 0;
    g_RxMessage.m_Checksum = 0;
    g_RxMessage.m_ReadBytes = 0;
    g_RxMessage.m_pData = RxDataBuffer;
    
    // The response data length depends on the category and setting
    if (Category == 0x80 && Setting == 0x02)
        // Firmware upgrade protocol 1 file part: Return message contains file-part
        // sequence number
        g_RxMessage.m_DataLength = 2;
    else if (Category == 0x85 && Setting == 0x03)
        // Firmware upgrade protocol 2 programming data: Return message contains no data
        g_RxMessage.m_DataLength = 0;
    else if (Category == 0x86 && Setting == 0x07)
        // Firmware upgrade protocol 3 programming data: Return message contains no data
        g_RxMessage.m_DataLength = 0;
    else
        // Otherwise, response message will be a copy of the command
        g_RxMessage.m_DataLength = NumBytes / 2;

    return DtmSendCommandWithRetry(0);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- GetFieldSize -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Returns the fieldsize in bytes for each datatype
//
int  GetFieldSize(int Category, int Setting)
{
    int  DtmSetting;

    // Get category and setting info
    for (DtmSetting=0; DtmSetting<g_NumDtmSettingInfoItems; DtmSetting++)
        if (g_DtmSettingInfo[DtmSetting].m_Category==Category 
                                      && g_DtmSettingInfo[DtmSetting].m_Setting==Setting)
            break;

    // Valid category and setting given?
    if (DtmSetting == g_NumDtmSettingInfoItems)
        return 0;

    // Return fieldsize
    switch (g_DtmSettingInfo[DtmSetting].m_DataType)
    {
    case D_UINT8:       // 8-bit unsigned integer
        return 1; 
        break;

    case D_UINT16:      // 16-bit unsigned integer
        return 2; 
        break;

    case D_INT32:       // 32-bit signed integer
    case D_UINT32:      // 32-bit unsigned integer
    case D_IP:          // IP address
        return 4; 
        break;

    case D_IPV6:        // IPv6 address
        return 16;
        break;

    case D_UINT64:      // 64-bit unsigned integer
        return 8; 
        break;

    case D_MAC:         // MAC address
        return 6;
        break;

    default:
        return 0;
        break;
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- IsCategoryWithIndex -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Helper function which returns true for categories with index parameter
//
int  IsCategoryWithIndex(int Category)
{
    int  CurrentCategory;

    // Get category info
    for (CurrentCategory=0; CurrentCategory<g_NumDtmCategoryInfoItems; CurrentCategory++)
        if (g_DtmCategoryInfo[CurrentCategory].m_Category==Category)
            break;

    // Valid category given?
    if (CurrentCategory == g_NumDtmCategoryInfoItems)
        return 0;

    // Return m_HasIndex
    return g_DtmCategoryInfo[CurrentCategory].m_HasIndex;
}

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Low-Level Helper Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmGetChecksum -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Calculates and returns the BCC over a number of bytes taken from a databuffer
//
int  DtmGetChecksum(char* pData, int Len, int StartValue)
{
    int  i;
    int  Checksum;

    Checksum = StartValue;
    for (i=0; i<Len; i++)
        Checksum += *(pData+i);

    Checksum = (~(Checksum & 0xFF)) + 1;
    return Checksum & 0xFF;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmHexToInt -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Convert a number of ASCII hex nibbles ('0'-'9','A'-'F') to an integer (max 64bit)
//
// return value is a 64-bit integer but depending on the 'Len' parameter can contain
// 1 to 8 bytes
//
unsigned long long  DtmHexToInt(char* pHex, int Len)
{
    unsigned long long  Data;
    char  Nibble, i;
    Nibble = 0;
    Data = 0;

    for (i=0; i<Len; i++)
    {
        if (*pHex>='0' && *pHex<='9')
            Nibble = *pHex - '0';
        else if (*pHex>='A' && *pHex<='F')
            Nibble = *pHex - 'A' + 10;
        else if (*pHex>='a' && *pHex<='f')
            Nibble = *pHex - 'a' + 10;
        Data = (Data<<4) + Nibble;
        pHex++;
    }
    return Data;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmHexToByte -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Parses a ASCII hex buffer into a binary data buffer
//
// Parameters:
// -pHex        : Pointer to input buffer to convert
// -Len         : Number of ASCII characters to parse
// -pReturnData : Pointer to least significant part of buffer where binary data is written
// -InputDataIsMsb    : Indicates if pHex is pointing to most or least significant data
//
// Return value is the number of binary bytes that were written
//
int  DtmHexToByte(char* pHex, int Len, char* pReturnData, int InputDataIsMsb)
{
    int  i;

    if (!InputDataIsMsb && Len!=1)
        // Loop from the least significant ASCII Hex data to the most significant part
        pHex += Len-2;
    for (i=0; i<(Len/2); i++)
    {
        *pReturnData = (char)(DtmHexToInt(pHex,2));
        pReturnData++;
        if (!InputDataIsMsb)
            pHex -= 2;
        else
            pHex += 2;
    }
    if (Len%2)
        *pReturnData = (char)(DtmHexToInt(pHex,1));

    return (Len/2) + (Len%2);
}


//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmByteToHex -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Converts a binary data buffer into an ASCII Hex data buffer
//
// Parameters:
// -pByte : Pointer to buffer containing binary data to parse
// -Len   : Number of binary bytes to parse
// -pHex  : Pointer to buffer where the least significant part of the ASCII data is stored
//          (pHex[0] = least significant part, pHex[Len*2] = most significant part)
// -InputDataIsMsb    : Indicates if pHex is pointing to most or least significant data
//
// Return value is the number of ASCII characters that were generated
//
int  DtmByteToHex(unsigned char* pByte, int Len, char* pHex, int InputDataIsMsb)
{
    char  Nibble;
    int  i;
    
    if (!InputDataIsMsb)
        // Loop from most significant part to least significant part
        pByte += Len-1;

    for (i=0; i<Len; i++)
    {
        Nibble = (*pByte>>4) & 0x0F;
        if (Nibble>=0 && Nibble<=9)
            *pHex = '0' + Nibble;
        else
            *pHex = 'A' + Nibble - 10;
        pHex++;

        Nibble = *pByte & 0x0F;
        if (Nibble>=0 && Nibble<=9)
            *pHex = '0' + Nibble;
        else
            *pHex = 'A' + Nibble - 10;
        pHex++;

        if (!InputDataIsMsb)
            pByte--;
        else
            pByte++;
    }
    return (Len*2);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmIntToHex -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Converts an integer into a number of ASCII characters
//
void  DtmIntToHex(int Input, int Bytes, char* pHex)
{
    char  Nibble;
    int  i;

    pHex += Bytes*2-1;
    for (i=0; i<(Bytes*2); i++)
    {
        Nibble = Input & 0xF;
        Input >>= 4;
        if (Nibble>=0 && Nibble<=9)
            *pHex = '0' + Nibble;
        else              
            *pHex = 'A' + Nibble - 10;
        pHex--;
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- BinToAsc128 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Convert a buffer containing binary data to ASCII characters 128 to 255
//
int  BinToAsc128(unsigned char* pDataIn, int LengthIn, char* pDataOut)
{
    int  BytesToDo, i, j;
    BytesToDo = LengthIn;

    while (BytesToDo > 0)
    {
        if (BytesToDo >= 7)
        {
            *pDataOut++ = 0x80 + (*pDataIn & 0x7F);
            for (i=1; i<7; i++)
            {
                j = i - 1;
                *pDataOut++ = 0x80 + ((*(pDataIn + i) & 0x7F >> i) << i) +
                                               ((*(pDataIn + j) & ~(0x7F >> j)) >> (7-j));
            }
            *pDataOut++ = 0x80 + ((*(pDataIn + 6) & 0xFE) >> 1);
            BytesToDo -= 7;
            pDataIn += 7;
        } else {
            i = 1;
            j = 0;
            *pDataOut++ = 0x80 + (*pDataIn & 0x7F);
            while (BytesToDo > 1)
            {
                *pDataOut++ = 0x80 + ((*(pDataIn + i) & 0x7F >> i) << i) +
                                               ((*(pDataIn + j) & ~(0x7F >> j)) >> (7-j));
                BytesToDo--;
                i++;
                j++;
            }
            *pDataOut++ = 0x80 + ((*(pDataIn + j) & ~(0x7F >> j)) >> (7-j));
            BytesToDo--;
        }
    }
    *pDataOut = '\0';

    if (LengthIn*8%7 > 0)
        return (LengthIn*8/7 + 1);

    return (LengthIn*8/7);
}
