//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtmHandler.h *#*#*#*#*#*#*#*# (C) 2012-2019 DekTec
//
// DtmHandler - Protocol handler for DTM-32xx devices - Type definitions and declarations
//
// The open source "DTM handler" is a set of definitions and functions that can help you 
// build a protocol handler for DekTec's DTM-32xx devices. This C source code can be 
// compiled and linked into a C or C++ application. Please refer to DtmCmd for an open
// source example of a DTM-32xx command-line control program that uses the DTM handler.
//
// This header file contains the definitions and declarations that are needed by users
// of the DTM handler. Definitions and declarations that are used internally in the DTM
// handler only are specified in DtmHandler.c

//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ LICENSE +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

// Copyright (C) 2012-2019 DekTec Digital Video B.V.
//
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
//  1. Redistributions of source code must retain the above copyright notice, this list
//     of conditions and the following disclaimer.
//  2. Redistributions in binary format must reproduce the above copyright notice, this
//     list of conditions and the following disclaimer in the documentation.
//  3. The source code may not be modified for the express purpose of enabling hardware
//     features for which no genuine license has been obtained.
//
// THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DEKTEC DIGITAL VIDEO BV, ITS AGENTS OR ITS EMPLOYEES BE LIABLE FOR
// ANY DIRECT, INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR OTHER DAMAGES (INCLUDING DAMAGES
// FOR THE LOSS OF USE, INFORMATION, GOODWILL, PROFIT, WORK STOPPAGE, DATA, BUSINESS OR
// REVENUE) UNDER ANY CIRCUMSTANCES, OR UNDER ANY LEGAL THEORY, WHETHER IN CONTRACT, IN
// TORT, IN NEGLIGENCE, OR OTHERWISE, ARISING FROM THE USE OF, OR INABILITY TO USE THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Usage of DTM Handler -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// In order to use functions of the DTM handler for communicating with DTM devices, you
// have to supply the following functions, which are used to implement platform- and 
// application-specific functionality. These functions are declared in this header file.
//
// Required callback functions:
// - DtmRxFunction              Receive data from I2C / RS-232/422/485 port
// - DtmTxFunction              Transmit data to I2C / RS-232/422/485 port
// - DtmSleepFunction           Suspend execution for a given period of time
// 
// Optional callback functions:
// - DtmPumpFunction            Process GUI events to keep the GUI responsive
// - DtmUpgGetData              Read data from a firmware file
// - DtmUpgReportProgress       Report progress of the firmware upgrading process
//
// Call DtmHandlerInit to supply the pointers to these callback functions.
//
// The DTM handler provides the following communication functions:
// - DtmReadBytes               Read 1 or more bytes from any setting
// - DtmReadInt                 Read 1, 2 or 4 bytes from a compatible setting
// - DtmReadInt64               Read 8 bytes from a compatible setting
// - DtmWriteBytes              Write 1 or more bytes to any setting
// - DtmWriteInt                Write 1, 2 or 4 bytes to a compatible setting
// - DtmWriteInt64              Write 8 bytes to a compatible setting
// - DtmFirmwareUpgrade         Upgrade the firmware of a DTM-32xx device
//
// The DTM handler allocates several buffers for internal use. To minimize the use of 
// statically allocated memory you can define USE_DYN_UPG_BUFFER, which lets the DTM
// handler use malloc() and free() to manage allocation and deletion of buffers.
//
// The DTM handler uses the following data types:
// - char                       Required to be 8-bit
// - int                        Required to be 32-bit
// - unsigned long long         Required to be 64-bit
//
// Changelog:
// v1.1.1 - First version of the DTM handler released to the field
// v1.2.0 - Added Busy error code
//        - Added DtmWaitForReady function
//        - Using DtmWaitForReady after firmware erase for Upgrade Type 2 devices
//        - Changed receiver category's names and grouping
//        - Added several receiver statistics for DTM-3237
//        - Incremented max message timeout for firmware upgrade category 0x85
// v1.2.1 - Bugfix for firmware upgrade type 1 devices (e.g. DTM-3200), upgrading
//          from firmware v1.x to v2.0 was not possible
// v1.3.0 - Added RF level calibration setting for RF devices (e.g. DTM-3237)
// v1.4.0 - Added RO "All in category" setting 0 to category 0x01, 0x02, 0x90, 0x91, 0x92
// v1.5.0 - Added "Input port select" setting 0x07 to category 0x83
// v1.6.0 - Added "Satellite Info" category 0x94
// v1.7.0 - Fixed byte order of DtmWriteBytes and DtmReadBytes functions. The first byte 
//          read/writen is stored/retrieved from pData[0] in stead of pData[NumBytes-1]
// v1.8.0 - Categories and settings to support DTM-3224 added.
//          Support for IPv6 addresses added.
// v1.8.1 - Minor changes to DTM-3224 manageable items:
//          'IP Version' setting 0x0E category 0x82 was removed.
//          'IPv6 Address' moved from setting 0x0F to 0x0E.
//          'FEC dimension' moved from setting 0x10 to 0x0F.
// v1.9.0 - Improved error handling in DtmSendCommand. Protocol errors were the data is 
//          removed from the message are now also catched.
// v1.10.0- Implemented firmware upgrade for upgrade type 3 (e.g. DTM-3224)
// v1.11.0- Internal housekeeping, no functional changes
// v1.11.1- Don't retry sending the reboot command when transmit succeeded
// v1.12.0- Bugfix for parsing response messages: don't parse index as read/write field
// v1.12.1- Add minimal check to make sure we're not writing an overflowed value
// v1.13.0- Added settings "Tuner mode" (0x0E) and "Tuner bandwidth" (0x0F) to 
//          category "Receiver Settings" (0x90)
// v1.14.0- Added settings "Link state" (0x0D) ,"Tx count" (0x0E) and "Rx count" (0x0F) to 
//          category "Network Settings" (0x03)
//

#ifndef _DTMHANDLER_H
#define _DTMHANDLER_H

#ifdef __cplusplus
extern "C"{
#endif

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- DtmHandler Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Version consists of three numbers:            // Changes when:
#define DTMHANDLER_VERSION_MAJOR    1            //  Major function(s) added or changed
#define DTMHANDLER_VERSION_MINOR    14           //  Minor function(s) added or changed
#define DTMHANDLER_VERSION_BUGFIX   0            //  Bugs are fixed, no change in function
    
//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Helper Types -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Type of serial communication protocol used to control the DTM-32xx.
// The communication type determines details of packaging and interpreting messages. 
typedef enum _DtmComType
{
    I2C,                            // Control via I2C bus
    UART                            // Control via RS-232/422/485
}  DtmComType;

// Data types used in reading/writing settings
typedef enum _DtmDataTypes
{
    D_UINT8,                        // 8-bit unsigned integer
    D_UINT16,                       // 16-bit unsigned integer
    D_INT32,                        // 32-bit signed integer
    D_UINT32,                       // 32-bit unsigned integer
    D_UINT64,                       // 64-bit unsigned integer
    D_IP,                           // IP address
    D_MAC,                          // MAC address
    D_IPV6                          // IPv6 address
} DtmDataTypes;

// Result type for DTM handler operations
typedef enum _DtmResult
{
    DTM_OK = 0,
    DTM_E_ERROR_REPLY,
    DTM_E_INCOMPLETE_MSG,
    DTM_E_INCORRECT_CHECKSUM,
    DTM_E_INCORRECT_REPLY,
    DTM_E_INVALID_PARAMETER,
    DTM_E_INVALID_MODE,
    DTM_E_NOT_SUPPORTED,
    DTM_E_NULL_PNTR,
    DTM_E_OUT_OF_MEM,
    DTM_E_PORT_CLOSED,
    DTM_E_RX_FAILED,
    DTM_E_RXFUNC_NULL_PNTR,
    DTM_E_TIME_OUT,
    DTM_E_TX_FAILED,
    DTM_E_TXFUNC_NULL_PNTR,
    DTM_E_UPG_FAILED,
    DTM_E_UPG_FILE_ERROR,
    DTM_E_UPG_FILE_INVALID_CHECKSUM,
    DTM_E_UPG_NULL_PNTR,
    DTM_E_UPG_VERIFY_FAILED,
    DTM_E_BUSY_REPLY,
    DTM_GOT_LESS_DATA,
    DTM_GOT_MORE_DATA,
    DTM_GOT_NO_DATA,
    DTM_REPLY_IGNORED,
    DTM_OUT_OF_RANGE
} DtmResult;

// Upgrade phase, used by ReportProgress callback to report progress of upgrading
// the firmware of a DTM-32xx
typedef enum _DtmUpgPhase 
{ 
    INIT,                           // Preparing the firmware upgrade
    UPLOADING,                      // Uploading the firmware data to the DTM-32xx
    PROGRAMMING,                    // Programming the EEPROM
    FINISHED                        // Programming done
} DtmUpgPhase;

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ DTM Handler API Functions +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Main Callback Functions -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

// Function to receive serial data (I2C or RS-232/422/485) from the DTM-32xx
typedef DtmResult  (*DtmRxFunction)(void* pOpaque, int DtmAddr, char* pData,
                                               int Expected, int* pReturned, int TimeOut);

// Function to transmit serial data (I2C or RS-232/422/485) to the DTM-32xx
typedef DtmResult  (*DtmTxFunction)(void* pOpaque, int DtmAddr, char* pData, int Length);

// Function to suspend the execution until the time-out interval elapses
typedef DtmResult  (*DtmSleepFunction)(void* pOpaque, int Milliseconds);

// Function to keep the GUI alive. It is called repeatedly whenever the DTM handler code
// is waiting. Can be set to (DtmPumpGuiFunction) NULL if you have no GUI.
typedef void  (*DtmPumpFunction)(void* pOpaque);

//-.-.-.-.-.-.-.-.-.-.-.- Callback Functions for Firmware Upgrade -.-.-.-.-.-.-.-.-.-.-.-.

// Function to read data from a firmware file
typedef DtmResult  (*DtmUpgGetData)(void* pOpaque, char* pData, int NumToRead);

// Function to report the progress of the firmware upgrading process
typedef void  (*DtmUpgReportProgress)(void* pOpaque, DtmUpgPhase Phase, int Progress);

//-.-.-.-.-.-.-.-.-.- Initialisation / Read / Write / Firmware Upgrade -.-.-.-.-.-.-.-.-.-

// DTM handler initialization. Use this function to set the appropriate communication
// type (UART or I2C) and to hook up the platform-specific communication functions
void  DtmHandlerInit(void* pOpaque, DtmComType ComType, DtmRxFunction pRxFunction, 
                               DtmTxFunction pTxFunction, DtmSleepFunction pSleepFunction, 
                               DtmPumpFunction pPumpFunction);

// Read bytes (1..1000 bytes) from DTM-32xx
DtmResult  DtmReadBytes(int DtmAddr, int Category, int Setting, int Index, char* pData, 
                                                                            int NumBytes);

// Read integer (1..4 bytes) from DTM-32xx
DtmResult  DtmReadInt(int DtmAddr, int Category, int Setting, int Index, int* pValue);

// Read 64-bit integer (8 bytes) from DTM-32xx
DtmResult  DtmReadInt64(int DtmAddr, int Category, int Setting, int Index, 
                                                              unsigned long long* pValue);

// Write bytes (1..1000 bytes) to DTM-32xx
DtmResult  DtmWriteBytes(int DtmAddr, int Category, int Setting, int Index, char* pData, 
                                                                            int NumBytes);

// Write integer (1..4 bytes) to DTM-32xx
DtmResult  DtmWriteInt(int DtmAddr, int Category, int Setting, int Index, int pValue);

// Write 64-bit integer (8 bytes) to DTM-32xx
DtmResult  DtmWriteInt64(int DtmAddr, int Category, int Setting, int Index, 
                                                               unsigned long long pValue);

// Get the method of upgrading firmware (the "upgrade type") for this DTM-32xx device 
// Upgrade type is 1 for DTM-3200 and 2 for DTM-3237
DtmResult  DtmGetUpgradeType(int DtmAddr, int* pUpgradeType);

// Upgrade the firmware of the DTM-32xx
// *pGetData is used to read data from the firmware file
DtmResult  DtmFirmwareUpgrade(int DtmAddr, int FileSize, DtmUpgGetData pGetData, 
                                                    DtmUpgReportProgress pReportProgress);

// This function waits for the device to be ready by reading the "Busy" setting
DtmResult  DtmWaitForReady(int DtmAddr, int IgnoreErrors);

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Internal helper functions -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Read bytes (1..1000 bytes) from DTM-32xx
DtmResult  DtmReadBytesInt(int DtmAddr, int Category, int Setting, int Index, 
                            char* pTxData, int TxNumBytes, char* pRxData, int RxNumBytes);

//-.-.-.-.-.-.-.-.-.-.-.-.- DTM-32xx Protocol Description Tables -.-.-.-.-.-.-.-.-.-.-.-.-

// The following data structures and lookup tables describe the DTM protocol categories
// and settings. Please refer to the DTM-32xx manuals for a description of the DTM 
// control protocol.

// Structure to describe a settings category
typedef struct _DtmCategoryInfo
{
    int  m_Category;                // Integer identifying this category
    int  m_HasIndex;                // Settings in this category use index field yes/no
    int  m_MaxMsgTimeOut;           // Worst-case response time (ms)
    const char*  m_pName;           // Human readable string for this category
} DtmCategoryInfo;

// Table providing info per category (table is defined in DtmHandler.c)
extern const DtmCategoryInfo  g_DtmCategoryInfo[];
extern const int  g_NumDtmCategoryInfoItems;

// Structure to describe a DTM-32xx setting
typedef struct _DtmSettingInfo
{
    int  m_Category;                // Category to which this setting belongs
    int  m_Setting;                 // Integer identifying this setting
    int  m_Units;                   // 10 => Setting is specified in units of 0.1
    DtmDataTypes  m_DataType;       // Data type of this setting
    const char*  m_pName;           // Human readable string for this setting
} DtmSettingInfo;

// Table providing info per setting (table is defined in DtmHandler.c)
extern const DtmSettingInfo  g_DtmSettingInfo[];
extern const int  g_NumDtmSettingInfoItems;

// Structure to indicates for a DTM-32xx device whether the subtype setting is available
typedef struct _DtmDeviceSubTypePresence
{
    int  m_Type;                    // Device type, 3200 for DTM-3200
    int  m_SubTypePresence;         // Does this device have subtypes: 0=no, 1=yes
} DtmDeviceSubTypePresence;

// Table that indicates whether the subtype setting is available
extern const DtmDeviceSubTypePresence  g_DtmDeviceSubTypePresence[];
extern const int  g_NumDtmDeviceSubTypePresence;

// Structure to indicate the firmware-upgrade type for a DTM-32xx device
typedef struct _DtmUpgradeType
{
    int  m_Type;                    // Device type, e.g. 3200 for DTM-3200
    int  m_UpgradeType;             // Upgrade type, e.g. DTM-3200 is upgrade type 1 and
                                    // DTM-3237 is upgrade type 2
} DtmUpgradeType;

// Table that specifies the firmware-upgrade type for a DTM-32xx device
extern const DtmUpgradeType  g_DtmUpgradeTypes[];
extern const int  g_NumDtmUpgradeTypes;

// Structure to define the character encoding to send a firmware file to a DTM-32xx
typedef struct _DtmFwFileCharEncoding
{
    int  m_Type;                    // Device type, e.g. 3237 for DTM-3237
    int  m_FirmwareVersion;         // Firmware version, e.g. 10 for v1.0
    int  m_BootloaderVersion;       // Bootloader version, e.g. 1 for v1.0
    int  m_UseAsc128Encoding;       // 0 = normal ASCII encoding, 1 = ASC128 encoding
} DtmFwFileCharEncoding;

// Table that defines the firmware character encoding per DTM-32xx / version
extern const DtmFwFileCharEncoding  g_DtmFwFileCharEncoding[];
extern const int  g_NumDtmFwFileCharEncoding;

#ifdef __cplusplus
}
#endif

#endif
